import json
import os
import requests
import tempfile
import yaml
import ConfigParser

from ansible.errors import AnsibleError as ae
from ansible.plugins.action import ActionBase
from ansible.utils.boolean import boolean
from requests.auth import HTTPBasicAuth
try:
    from __main__ import display
except ImportError:
    from ansible.utils.display import Display
    display = Display()

class ActionModule(ActionBase):
    '''Manage hosts and services monitored at uptime'''

    BYPASS_HOST_LOOP = True
    TRANSFERS_FILES = False

    def __init__(self, task, connection, play_context, loader, templar,
                 shared_loader_obj):
        super(ActionModule, self).__init__(task, connection, play_context,
                                           loader, templar, shared_loader_obj)
        self._supports_check_mode = False
        self.args = {}
        self.vars = {}

        self.pollerParams = [
            'http_options',
            'match',
            'hipchat_token',
            'hipchat_roomid',
            'mattermost_hook',
            'tracker_source',
            'tracker_timeshift',
            'tracker_language',
            'tracker_hidemine',
            'tracker_username',
            'tracker_password',
            'tracker_project_include',
            'tracker_project_exclude'
        ]


    def run(self, tmp=None, task_vars=None):
        if task_vars is None:
            task_vars = dict()
        self.vars = task_vars

        result = super(ActionModule, self).run(tmp, task_vars)

        if self._play_context.check_mode:
            result['skipped'] = True
            result['msg'] = 'check mode not supported for this module'
            return result

        just_download = boolean(self._task.args.get('readonly'))

        configPath = '/etc/uptime.cfg'
        if not os.path.exists(configPath):
            result['msg'] = 'Uptime not configured'
            result['changed'] = False
            return result
        self.config = ConfigParser.ConfigParser()
        self.config.read(configPath)
        section = 'Credentials'
        if self.config.has_option(section, 'url') and self.config.has_option(section, 'username') and self.config.has_option(section, 'password'):
            self.url = self.config.get(section, 'url')
            self.username = self.config.get(section, 'username')
            self.password = self.config.get(section, 'password')
        else:
            raise ae('Can not find credentials')

        display.vv('Load checks from Uptime')
        self.checks = self._request('checks')

        if just_download:
            output = self._task.args.get('output', tempfile.mktemp(prefix='uptime_', suffix='.yml'))
            with open(output, 'w') as output_file:
                yaml.safe_dump(self.checks, output_file, encoding='utf-8', allow_unicode=True)
            result['msg'] = 'Downloaded settings to %s' % output
            result['changed'] = False
            return result


        if not self.config.has_section('Domains'):
            raise ae('Can not find domain settings')

        options = self.config.options('Domains')

        if self.config.has_section('Paths'):
            paths = self.config.options('Paths')
        else:
            paths = {}

        if self.config.has_section('Matches'):
            matches = self.config.options('Matches')
        else:
            matches = {}

        display.vv('Process hosts')
        for host in self.vars.get('ansible_play_hosts'):
            display.vv('- ' + host)
            self._host(host, options, paths, matches)

        result['msg'] = 'Completed successfully!'
        return result


    def _request(self, path, data = None, method = 'GET'):
        encoder = json.JSONEncoder(encoding=None)
        postData = {}

        if data:
            if method == 'GET':
                method = 'POST'

            for key in data:
                item = data.get(key)
                if type(item) is list:
                    item = ','.join(item)
                if type(item) is dict:
                    item = encoder.encode(item)
                if type(item) is int:
                    item = str(item)
                if type(item) is bool:
                    item = str(item).lower()

                postData.__setitem__(key, unicode(item))

        auth = HTTPBasicAuth(self.username, self.password)

        request_result = {}
        try:
            if method == 'GET':
                request_result = requests.get(self.url + '/api/' + path, auth = auth)
            elif method == 'POST':
                request_result = requests.post(self.url + '/api/' + path, auth = auth, data = postData)
            elif method == 'PUT':
                request_result = requests.put(self.url + '/api/' + path, auth = auth, data = postData)
            elif method == 'DELETE':
                request_result = requests.delete(self.url + '/api/' + path, auth = auth)
        except ae, e:
            raise ae('No result from Uptime API')

        display.vvvvvv(request_result)
        if request_result.status_code != 200:
            msg = 'Error %s' % request_result.status_code
            raise ae('%s' % msg)

        decoder = json.JSONDecoder()
        if request_result.content == '':
            content = []
        else:
            content = decoder.decode(request_result.content)
        return content


    def _cleanupCheck(self, check):
        check['interval'] = check['interval'] / 1000
        return check


    def _createCheck(self, check):
        display.vv('    Creating %s' % (check.get('url')))
        self._cleanupCheck(check)
        self._request('checks', check, 'PUT')


    def _deleteCheck(self, check):
        display.vv('    Deleting %s' % (check.get('url')))
        self._request('checks/' + check.get('_id'), None, 'DELETE')


    def _updateCheck(self, existing, check):
        changed = False

        if 'pollerParams' not in existing:
            existing['pollerParams'] = {}

        for key, value in check.iteritems():
            if key == 'enabled':
                continue
            if key in self.pollerParams:
                existing[key] = existing['pollerParams'].get(key, '')
                if existing[key] is None:
                    existing[key] = ''
            elif key not in existing:
                existing[key] = None

            if existing[key] != check[key]:
                display.vv('        %s changed: "%s" vs "%s"' % (key, existing[key], check[key]))
                existing[key] = check[key]
                changed = True

        if changed:
            display.vv('    Updating %s' % (existing.get('url')))
            existing = self._cleanupCheck(existing)
            self._request('checks/' + existing.get('_id'), existing, 'POST')


    def _findExistingCheck(self, url, type, tracker_source, tracker_project_include):
        for check in self.checks:
            pollerParams = check.get('pollerParams', [])
            if (check.get('url') == url and check.get('type') == type and pollerParams.get('tracker_source', '') == tracker_source and pollerParams.get('tracker_project_include', '') == tracker_project_include):
                return check
        return False


    def _buildUrl(self, item, field, path):
        authPrefix = ''
        auth = item.get('apache_auth')
        if auth and auth.get('type') == 'Basic':
            authPrefix = auth.get('user') + ':' + auth.get('password') + '@'
        url = item.get('protocol', 'https') + '://' + authPrefix + item.get(field)
        uptime = item.get('uptime')
        if uptime and 'path' in uptime:
            path = uptime.get('path')
        if path and path != '':
            url += '/' + path
        return url


    def _defaultCheck(self):
        return {
            'enabled': True,
            'interval': 60000,
            'maxTime': 1500,
            'isPaused': False,
            'alertTreshold': 5,
        }


    def _addHostSettings(self, host_vars, section, check):
        settings = host_vars.get(section)
        if settings:
            for key in settings:
                item = section + '_' + key
                check[item] = settings[key]
                if not item in self.pollerParams:
                    self.pollerParams.append(item)
        return check


    def _buildCheck(self, host_vars, item, url):
        check = self._defaultCheck()
        check['url'] = url
        check['type'] = item.get('protocol', 'https')
        check = self._addHostSettings(host_vars, 'mattermost', check)
        uptime = item.get('uptime')
        if uptime:
            for key in self._defaultCheck():
                if key in uptime:
                    check[key] = uptime[key]
            if 'name' in uptime:
                check['name'] = uptime['name']
            if 'type' in uptime:
                check['type'] = uptime['type']
            if 'tags' in uptime:
                check['tags'] = uptime['tags']
            if 'pollerParams' in uptime:
                for param in self.pollerParams:
                    check[param] = uptime['pollerParams'].get(param, '')

        if 'tags' not in check:
            check['tags'] = ['none']

        return check


    def _host(self, host, options, paths, matches):
        host_vars = self.vars['hostvars'][host]
        for option in options:
            display.vv('  - ' + option)
            items = host_vars.get(option)
            if items:
                parts = self.config.get('Domains', option).split('.')
                field = parts.pop()
                if option in paths:
                    path = self.config.get('Paths', option)
                else:
                    path = ''
                if option in matches:
                    match = self.config.get('Matches', option)
                else:
                    match = ''
                self._part(host_vars, items, 0, parts, field, path, match)


    def _part(self, host_vars, items, level, parts, field, path, match):
        if len(parts) > level:
            next = parts[level]
            level += 1
            display.vv('    ' + next)
            for item in items:
                nextitems = item.get(next)
                if nextitems:
                    self._part(host_vars, nextitems, level, parts, field, path, match)
        else:
            for item in items:
                value = item.get(field)
                if value:
                    display.vv('      ' + value)
                    if match != '':
                        if not item.has_key('pollerParams'):
                            item['pollerParams'] = {}
                        item['pollerParams']['match'] = '/' + match + '/'
                    self._processItem(host_vars, item, field, path)


    def _processItem(self, host_vars, item, field, path):
        url = self._buildUrl(item, field, path)
        check = self._buildCheck(host_vars, item, url)
        tracker_source = check.get('tracker_source', '')
        tracker_project_include = check.get('tracker_project_include', '')
        existing = self._findExistingCheck(url, check['type'], tracker_source, tracker_project_include)
        if existing:
            if check['enabled']:
                self._updateCheck(existing, check)
            else:
                self._deleteCheck(existing)
        else:
            if check['enabled']:
                self._createCheck(check)

